@@ -179,7 +179,7 @@ cc_timezone_map_finalize (GObject *object)
 
   if (priv->tzdb)
     {
-      g_free (priv->tzdb);
+      tz_db_free (priv->tzdb);
       priv->tzdb = NULL;
     }
 
@@ -194,10 +194,8 @@ cc_timezone_map_size_request (GtkWidget      *widget,
 {
   CcTimezoneMapPrivate *priv = CC_TIMEZONE_MAP (widget)->priv;
 
-  req->width = gdk_pixbuf_get_width (priv->orig_background) / 2;
-  req->height = gdk_pixbuf_get_height (priv->orig_background) / 2;
-
-  GTK_WIDGET_CLASS (cc_timezone_map_parent_class)->size_request (widget, req);
+  req->width = gdk_pixbuf_get_width (priv->orig_background) * 0.6;
+  req->height = gdk_pixbuf_get_height (priv->orig_background) * 0.6;
 }
 
 static void
@@ -304,19 +302,16 @@ convert_latitude_to_y (gdouble latitude, gdouble map_height)
 
 
 static gboolean
-cc_timezone_map_expose_event (GtkWidget      *widget,
-                              GdkEventExpose *event)
+cc_timezone_map_draw (GtkWidget *widget,
+                      cairo_t   *cr)
 {
   CcTimezoneMapPrivate *priv = CC_TIMEZONE_MAP (widget)->priv;
   GdkPixbuf *hilight, *orig_hilight;
-  cairo_t *cr;
   GtkAllocation alloc;
   gchar *file;
   GError *err = NULL;
   gdouble pointx, pointy;
 
-  cr = gdk_cairo_create (gtk_widget_get_window (widget));
-
   gtk_widget_get_allocation (widget, &alloc);
 
   /* paint background */
@@ -327,6 +322,9 @@ cc_timezone_map_expose_event (GtkWidget      *widget,
   file = g_strdup_printf (DATADIR "/timezone_%g.png",
                           priv->selected_offset);
   orig_hilight = gdk_pixbuf_new_from_file (file, &err);
+  g_free (file);
+  file = NULL;
+
   if (!orig_hilight)
     {
       g_warning ("Could not load hilight: %s",
@@ -348,17 +346,92 @@ cc_timezone_map_expose_event (GtkWidget      *widget,
 
   if (priv->location)
     {
+      PangoLayout *layout;
+      PangoRectangle rect;
+      gint width, height;
+      gdouble x1, y1, radius = 5;
+      gchar *zone;
+
       pointx = convert_longtitude_to_x (priv->location->longitude, alloc.width);
       pointy = convert_latitude_to_y (priv->location->latitude, alloc.height);
 
+      if (pointy > alloc.height)
+        pointy = alloc.height;
+
+      /* allow for the line width */
+      pointy -= 2;
+
+      zone = g_strdup (priv->location->zone);
+
+      /* convert underscores to spaces */
+      g_strdelimit (zone, "_", ' ');
+
+      layout = gtk_widget_create_pango_layout (widget, zone);
+
+      pango_layout_get_pixel_extents (layout, NULL, &rect);
+      width = rect.width - rect.x;
+      height = rect.height - rect.y;
+
+      x1 = (int) (pointx - width / 2);
+      y1 = (int) (pointy - 10 - height - radius);
+
+      /* rotate the bubble upside-down if there is not enough vertical room */
+      if (y1 < radius)
+        {
+          cairo_translate (cr, pointx, pointy);
+          cairo_rotate (cr, G_PI);
+          cairo_translate (cr, -pointx, -pointy);
+        }
+
+      /* offset the arrow if there is not enough horizontal room */
+      if (x1 - radius - 2 < 0)
+        x1 -= (x1 - radius - 2);
+
+      if (x1 + width + radius + 2 > alloc.width)
+        x1 -= (x1 + width + radius + 2) - alloc.width;
+
+      /* draw the bubble */
+      cairo_arc (cr, x1, y1, radius, G_PI, G_PI * 1.5);
+      cairo_line_to (cr, x1 + width, y1 - radius);
+
+      cairo_arc (cr, x1 + width, y1, radius, G_PI * 1.5, 0);
+      cairo_line_to (cr, x1 + width + radius, y1 + height);
+
+      cairo_arc (cr, x1 + width, y1 + height, radius, 0, G_PI * 0.5);
+      cairo_line_to (cr, pointx + 10, pointy - 10);
+
+      cairo_line_to (cr, pointx, pointy);
+      cairo_line_to (cr, pointx - 10, pointy - 10);
+      cairo_line_to (cr, x1, y1 + height + radius);
+
+      cairo_arc (cr, x1, y1 + height, radius, G_PI * 0.5, G_PI);
+      cairo_line_to (cr, x1 - radius, y1);
+
+      cairo_set_source_rgba (cr, 1, 1, 1, 0.7);
+      cairo_fill_preserve (cr);
+
+      cairo_set_source_rgba (cr, 0, 0, 0, 0.7);
+      cairo_stroke (cr);
+
+
+      if (y1 < radius)
+        {
+          cairo_translate (cr, pointx, pointy);
+          cairo_rotate (cr, G_PI);
+          cairo_translate (cr, -pointx, -pointy);
+
+          y1 += height + 20 + 2 * radius;
+        }
+
+      /* draw the location name */
       cairo_set_source_rgb (cr, 0, 0, 0);
-      cairo_arc (cr, pointx, pointy, 3.0, 0, 2 * G_PI);
-      cairo_fill (cr);
+      cairo_move_to (cr, x1, y1);
+      pango_cairo_show_layout (cr, layout);
+
+      g_object_unref (layout);
+      g_free (zone);
     }
 
-  cairo_destroy (cr);
-
-
   return TRUE;
 }
 
@@ -379,7 +452,7 @@ cc_timezone_map_class_init (CcTimezoneMapClass *klass)
   widget_class->size_request = cc_timezone_map_size_request;
   widget_class->size_allocate = cc_timezone_map_size_allocate;
   widget_class->realize = cc_timezone_map_realize;
-  widget_class->expose_event = cc_timezone_map_expose_event;
+  widget_class->draw = cc_timezone_map_draw;
 
   signals[LOCATION_CHANGED] = g_signal_new ("location-changed",
                                             CC_TYPE_TIMEZONE_MAP,
@@ -561,4 +634,6 @@ cc_timezone_map_set_timezone (CcTimezoneMap *map,
           break;
         }
     }
+
+  gtk_widget_queue_draw (GTK_WIDGET (map));
 }
