@@ -23,7 +23,7 @@
 
 #include "cc-default-applications-panel.h"
 
-#include <glib/gi18n.h>
+#include <glib/gi18n-lib.h>
 
 void
 g_io_module_load (GIOModule *module)
